(*************************************************************************

ZMIXER v.2 - very simple Cyrillic string cipher with 16 char secret key.

Key may be any ASCII string from 1 to 16 characters.

Toy cipher by (C) 2009 Alexander Myasnikov

Web: www.darksoftware.narod.ru

*************************************************************************)


{$R-}

unit zmixer;

interface

function Decrypt (s: string): string;
function Crypt (s: string): string;
procedure SetKey (s: string);


implementation

const
  lowr = 192;

const
  highr = 255;

const
  modval = (highr - lowr) + 1;

var
  stable: array [lowr..highr] of byte;
  bstable: array [lowr..highr] of byte;
  stable2: array [lowr..highr] of byte;
  bstable2: array [lowr..highr] of byte;
  key: array [1..16] of byte;
  iv:  longword;


procedure SetKey (s: string);
var
  i, x, y, vec, max: longword;
  n: byte;
  was: array [lowr..highr] of byte;
  a: array [1..16] of longword;
begin

  x := 1;

  s := copy(s, 1, 16);

  for I := lowr to highr do
    was[i] := 0;


  if length(s) < 16 then
    begin
    for I := 0 to 15 - length(s) do
      s := s + s[(i mod length(s)) + 1];
    end;

  vec := 1;

  for I := 1 to 16 do
    begin
    a[i] := byte(s[I]);
    key[i] := (a[i] mod modval) + lowr;
    vec := vec + (a[i] * i);
    x := (x * (x + 1) shr 1) + a[i];
    end;

  iv := (vec mod modval) + lowr;


  for I := lowr to highr do
    begin

    repeat

      a[(x mod 16) + 1] := (a[1] + a[2]) + ((a[3] - a[4]) xor
        (a[5] shr (a[6] mod 8))) + (a[7] shl (a[8] mod 8)) +
        (a[6] xor a[8]) xor ((a[5] + x) xor (a[7] + i)) +
        (a[9] - a[10] + ((a[10] shl 6) xor (a[10] shr 11))) +
        (a[12] + a[13]) + (a[14] xor a[15]) + (a[16] xor (a[16] * 9)) +
        (a[1] + (a[1] xor x));

      n := (a[(x mod 16) + 1] mod modval) + lowr;
      Inc(x);


    until (was[n] = 0) and ((n <> i) or (i = highr));


    if (n = i) and (i = highr) then
      begin
      x := stable[highr - 1];
      stable[highr - 1] := n;
      bstable[n] := highr - 1;
      n := x;
      end;

    was[n] := 1;

    stable[i]  := n;
    bstable[n] := i;

    end;


  for I := lowr to highr do
    was[i] := 0;

  x := 1;

  for I := 1 to 16 do
    begin
    x := (x * (x + 1) shr 1) + a[i];
    end;

  for I := lowr to highr do
    begin

    repeat

      a[(x mod 16) + 1] := (a[1] + a[2]) + ((a[3] - a[4]) xor
        (a[5] shr (a[6] mod 8))) + (a[7] shl (a[8] mod 8)) +
        (a[6] xor a[8]) xor ((a[5] + x) xor (a[7] + i)) +
        (a[9] - a[10] + ((a[10] shl 6) xor (a[10] shr 11))) +
        (a[12] + a[13]) + (a[14] xor a[15]) + (a[16] xor (a[16] * 9)) +
        (a[1] + (a[1] xor x));

      n := (a[(x mod 16) + 1] mod modval) + lowr;
      Inc(x);


    until (was[n] = 0) and ((n <> i) or (i = highr));


    if (n = i) and (i = highr) then
      begin
      x := stable2[highr - 1];
      stable2[highr - 1] := n;
      bstable2[n] := highr - 1;
      n := x;
      end;

    was[n] := 1;

    stable2[i]  := n;
    bstable2[n] := i;

    end;

end;


function ffirst (s: string): integer;
var
  n: integer;
begin
  for n := 1 to length(s) do
    begin
    Result := n;
    if byte(s[n]) in [lowr..highr] then
      break;

    end;
end;


function crypt (s: string): string;
var
  i, idx, First: integer;
  nc, prv: byte;

begin
  Result := '';

  First := ffirst(s);
  idx := 0;
  for I := 1 to length(s) do
    begin

    if idx < 16 then
      Inc(idx)
    else
      idx := 1;


    if byte(s[i]) in [lowr..highr] then
      begin
      if i = First then
        prv := iv;
      nc := (((stable[((stable2[byte(s[i])] xor key[idx]) mod modval) + lowr]) xor
        prv) mod modval) + lowr;
      prv := nc;
      Result := Result + char(nc);
      end
    else
      Result := Result + s[i];

    end;

end;

function decrypt (s: string): string;
var
  i, idx, First: integer;
  nc, prv: byte;

  function fprv (index: integer): byte;
  begin
    repeat
      Result := byte(s[index]);
      Dec(index);
    until Result in [lowr..highr];
  end;

begin
  Result := '';

  First := ffirst(s);

  idx := 0;

  for I := 1 to length(s) do
    begin
    if idx < 16 then
      Inc(idx)
    else
      idx := 1;
    end;


  for I := length(s) downto 1 do
    begin
    if byte(s[i]) in [lowr..highr] then
      begin
      if i = First then
        prv := iv
      else
        prv := fprv(i - 1);
      nc := bstable2[(((bstable[(((byte(s[i]) - lowr) mod modval) xor prv)] - lowr) mod
        modval) xor key[idx])];
      Result := char(nc) + Result;
      end
    else
      Result := s[i] + Result;


    if idx > 1 then
      Dec(idx)
    else
      idx := 16;

    end;

end;


end.
